# Copyright 2013-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from reahl.tofu import Fixture
from reahl.tofu import expected
from reahl.tofu.pytestsupport import with_fixtures

from reahl.mailutil.mail import Mailer, MailMessage, InvalidEmailAddressException
from reahl.dev.mailtest import MailTester


# noinspection PyAttributeOutsideInit
class MailerFixture(Fixture):
    def set_up(self):
        super().set_up()

        self.valid_email_addresses = ['someone@home.co.za', 'something@somewhere.com', 'j@j.ca']
        self.invalid_email_addresses = ['somethingWithoutTheAt', '@somethingThatBeginsWithTheAT',
                                        'somethingThatEndsWithTheAt@', '@', None]

        self.mail_checker_started = False
        self.mail_recipients = None
        self.mail_message = None

        self.mailer = Mailer()

    def start_mail_checker(self):
        self.mail_tester = MailTester(self.mail_checker)
        self.mail_tester.start()
        self.mail_checker_started = True

    def mail_checker(self, email_message, mailfrom, rcpttos):
        self.mail_sender = email_message['From']
        self.mail_recipients = email_message['To'].split(', ')
        self.mail_subject = email_message['Subject']
        self.mail_message = email_message.get_payload()[0].get_payload()[0].get_payload()

    def tear_down(self):
        super().tear_down()
        if self.mail_checker_started:
            self.mail_tester.stop()


@with_fixtures(MailerFixture)
def test_validation(mailer_fixture):
    fixture = mailer_fixture
    with expected(InvalidEmailAddressException):
        MailMessage(fixture.invalid_email_addresses[0],
                    fixture.valid_email_addresses,
                    'Hi',
                    'Some Message')

    with expected(InvalidEmailAddressException):
        MailMessage(fixture.valid_email_addresses[0],
                    fixture.invalid_email_addresses,
                    'Hi',
                    'Some Message')


@with_fixtures(MailerFixture)
def test_sending(mailer_fixture):
    fixture = mailer_fixture

    fixture.start_mail_checker()

    mail_message = MailMessage(fixture.valid_email_addresses[0],
                               fixture.valid_email_addresses,
                               'Hi',
                               'Some Message',
                               charset='ascii')
    fixture.mailer.send_message(mail_message)
    assert fixture.mail_sender == fixture.valid_email_addresses[0]
    assert fixture.mail_recipients == fixture.valid_email_addresses
    assert fixture.mail_subject == 'Hi'
    assert fixture.mail_message == 'Some Message'



# Copyright 2013-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for handling ReStructured Text."""

import docutils.io
import docutils.core


class RestructuredText:
    """A chunk of ReStructured Text.
    
       :param rst_text: A string containing the actual restructured text represented by this instance.
    """
    def __init__(self, rst_text):
        self.rst_text = rst_text
        
    def as_HTML_fragment(self, header_start=1, report_level=6, halt_level=4):
        """Returns this RestructuredText formatted as an HTML fragment.
        
           :keyword header_start: The n of the top-level <hn> for top-level heading in this text.
           :keyword report_level: Reports ReST error messages at or higher than this level.
           :keyword halt_level:   Error messages above this level result in exceptions.

           .. versionadded:: 4.0
        """
        settings = {'initial_header_level': header_start,
                    'doctitle_xform': False,
                    'output_encoding': 'unicode',
                    'input_encoding': 'unicode',
                    'report_level': report_level,
                    'halt_level': halt_level}

        parts = docutils.core.publish_parts(
            source_class=docutils.io.StringInput, source=self.rst_text, source_path=None,
            reader=None, reader_name='standalone',
            parser=None, parser_name='restructuredtext',
            writer=None, writer_name='html',
            settings=None, settings_spec=None,
            settings_overrides=settings,
            config_section=None,
            enable_exit_status=None)
        return parts['fragment']





# Copyright 2013-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility classes for sending simple email messages.

Run 'reahl componentinfo reahl-mailutil' for configuration information.

""" 


import re
import smtplib
import logging
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from reahl.component.context import ExecutionContext
from reahl.mailutil.rst import RestructuredText


class InvalidEmailAddressException(Exception):
    pass
    

class MailMessage:
    """Represents an email message, with one part containing plain text (presumed to be ReST), and the other
       an HTML representation of the same message.
       
       :param from_address: The "from" email address
       :param to_addresses: A list of "to" email addresses
       :param subject: The subject line
       :param rst_message: A message, expressed as a string containing ReStructured Text
       :keyword charset: The charset of `rst_message`
    """
    EMAIL_RE = re.compile("^[^\s]+@[^\s]+\.[^\s]{2,4}$")
    def __init__(self, from_address, to_addresses, subject, rst_message, charset='utf-8'):
            self.validate_email_addresses(to_addresses+[from_address])
            self.from_address = from_address
            self.to_addresses = to_addresses
            self.subject = subject
            self.rst_text = rst_message
            self.message_root = MIMEMultipart('related')
            self.message_root['Subject'] = subject
            self.message_root['From'] = from_address
            self.message_root['To'] = ", ".join(to_addresses)
            self.message_root.preamble = 'This is a multi-part message in MIME format.'

            # Encapsulate the plain and HTML versions of the message body in an
            # 'alternative' part, so message agents can decide which they want to display.
            self.message_alternative = MIMEMultipart('alternative')
            self.message_root.attach(self.message_alternative)

            message_text = MIMEText(rst_message.encode(charset), 'plain', charset)
            self.message_alternative.attach(message_text)

            rst = RestructuredText(rst_message)
            message_text = MIMEText(rst.as_HTML_fragment().encode(charset), 'html', charset)
            self.message_alternative.attach(message_text)

    def as_string(self):
        """Returns the message as ASCII-encoded string for sending."""
        return self.message_root.as_string()

    def validate_email_addresses(self, list_of_addresses):
        for address in list_of_addresses:
            if self.EMAIL_RE.search(address) is None:
                raise InvalidEmailAddressException(address)


class Mailer:
    """A proxy for a remote SMTP server.
    
       :keyword smtp_host: The host to connect to.
       :keyword smtp_port: The port to connect to.
       :keyword smtp_user: The username to use (if specified) for authentication.
       :keyword smtp_password: The password to authenticate with the smtp host.
       :keyword smtp_use_initial_encrypted_connection: If True, connect to the server using a secure connection (use with smtps)
       :keyword smtp_upgrade_initial_connection_to_encrypted: If True, connects to the server unencrypted, but then upgrade to a secure connection using STARTTLS (use with submission or smtp)
       :keyword smtp_keyfile: Keyfile to use for identifying the local end of the connection.
       :keyword smtp_certfile: Certfile to use for identifying the local end of the connection.
    """
    @classmethod
    def from_context(cls):
        """Returns a Mailer, using the system configuration."""
        config = ExecutionContext.get_context().config
        smtp_host = config.mail.smtp_host
        smtp_port = config.mail.smtp_port
        smtp_user = config.mail.smtp_user
        smtp_password = config.mail.smtp_password
        smtp_use_initial_encrypted_connection = config.mail.smtp_use_initial_encrypted_connection
        smtp_upgrade_initial_connection_to_encrypted = config.mail.smtp_upgrade_initial_connection_to_encrypted
        smtp_keyfile = config.mail.smtp_keyfile
        smtp_certfile = config.mail.smtp_certfile
        return cls(smtp_host=smtp_host, smtp_port=smtp_port, smtp_user=smtp_user, smtp_password=smtp_password,
                   smtp_use_initial_encrypted_connection=smtp_use_initial_encrypted_connection,
                   smtp_upgrade_initial_connection_to_encrypted=smtp_upgrade_initial_connection_to_encrypted,
                   smtp_keyfile=smtp_keyfile, smtp_certfile=smtp_certfile)


    def __init__(self, smtp_host='localhost', smtp_port=8025, smtp_user=None, smtp_password=None,
                 smtp_use_initial_encrypted_connection=False, 
                 smtp_upgrade_initial_connection_to_encrypted=False,
                 smtp_keyfile=None, smtp_certfile=None):
        self.smtp_host = smtp_host
        self.smtp_port = smtp_port
        self.smtp_user = smtp_user
        self.smtp_password = smtp_password
        self.smtp_use_initial_encrypted_connection = smtp_use_initial_encrypted_connection
        self.smtp_upgrade_initial_connection_to_encrypted = smtp_upgrade_initial_connection_to_encrypted
        self.smtp_keyfile = smtp_keyfile
        self.smtp_certfile = smtp_certfile

        self.smtp_server = None
        self.connected = False

    def connect(self):
        if self.smtp_use_initial_encrypted_connection:
            self.smtp_server = smtplib.SMTP_SSL(self.smtp_host, self.smtp_port,
                                                keyfile=self.smtp_keyfile, certfile=self.smtp_certfile)
        else:
            self.smtp_server = smtplib.SMTP(self.smtp_host, self.smtp_port)
            if self.smtp_upgrade_initial_connection_to_encrypted:
                self.smtp_server.ehlo()
                self.smtp_server.starttls(keyfile=self.smtp_keyfile, certfile=self.smtp_certfile)
                self.smtp_server.ehlo()
        if self.smtp_user:
            self.smtp_server.login(self.smtp_user, self.smtp_password)
        self.connected = True
        logging.debug('connected to server')

    def send_message(self, message):
        """Sends `message` (a :class:`MailMessage`) to the connected SMTP server."""
        try:
            self.connect()
            self.smtp_server.sendmail(message.from_address, message.to_addresses, message.as_string())
        finally:
            self.disconnect()

    def disconnect(self):
        if self.connected:
            try:
                self.smtp_server.quit()
            except smtplib.SMTPServerDisconnected as e:
                pass

        self.connected = False
        self.smtp_server = None





# Copyright 2013-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from reahl.component.config import Configuration, ConfigSetting


class MailConfig(Configuration):
    filename = 'mailutil.config.py'
    config_key = 'mail'
    smtp_host = ConfigSetting(default='localhost', description='The hostname used for sending SMTP email')
    smtp_port = ConfigSetting(default=8025, description='The port used for sending SMTP email', dangerous=True)
    smtp_user = ConfigSetting(default=None, description='The user name to connect to SMTP server')
    smtp_password = ConfigSetting(default=None, description='The password for the user to connect to the SMTP server')
    smtp_use_initial_encrypted_connection = ConfigSetting(default=False, description='If True, connect to the server using a secure connection (use with smtps)')
    smtp_upgrade_initial_connection_to_encrypted = ConfigSetting(default=False, description='If True, connects to the server unencrypted, but then upgrade to a secure connection using STARTTLS (use with submission or smtp)')
    smtp_keyfile = ConfigSetting(default=None, description='Keyfile to use for identifying the local end of the connection')
    smtp_certfile = ConfigSetting(default=None, description='Certfile to use for identifying the local end of the connection')



from setuptools import setup, Command
class InstallTestDependencies(Command):
    user_options = []
    def run(self):
        import sys
        import subprocess
        if self.distribution.tests_require: subprocess.check_call([sys.executable, "-m", "pip", "install", "-q"]+['pytest>=3.0', 'reahl-tofu', 'reahl-stubble', 'reahl-dev'])

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

setup(
    name='reahl-mailutil',
    version='6.1.0',
    description='Simple utilities for sending email from Reahl.',
    long_description='Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nReahl-mailutil is a simple library for sending emails (optionally from ReST sources). ',
    url='http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.mailutil', 'reahl.mailutil_dev'],
    py_modules=[],
    include_package_data=False,
    namespace_packages=['reahl'],
    install_requires=['reahl-component>=6.1,<6.2', 'docutils>=0.14,<0.19.999', 'Pygments>=2.1.0,<2.13.999'],
    setup_requires=['pytest-runner', 'reahl-component-metadata', 'setuptools >= 51.0.0, <= 62.1.0', 'setuptools-git >= 1.1', 'toml', 'wheel'],
    tests_require=['pytest>=3.0', 'reahl-tofu', 'reahl-stubble', 'reahl-dev'],
    extras_require={},
    cmdclass={'install_test_dependencies': InstallTestDependencies}
)
